class BetfairError(Exception):
    """Base class for all Exceptions in this package. Allow to hand in custom data with keyword arguments."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.__dict__.update(kwargs)


class JSONError(BetfairError):
    """Issue with the data serialization."""


class APIError(BetfairError):
    """Most general API error."""


class APINGException(APIError):
    """Errors thrown from a betfair operation."""


class AccountAPINGException(APIError):
    """An issue thrown from the accounts API."""


class IdentityError(APIError):
    """An issue thrown from the identity API."""


class LoginImpossible(IdentityError):
    """This probably needs manual resolution."""


class StreamError(APIError):
    """Issues related with the streaming API."""


class StreamAuthenticationError(StreamError):
    """Authentication for the stream API is not possible."""



from betfair_parser.spec.betting.enums import (
    MarketProjection,
    MarketSort,
    MatchProjection,
    OrderProjection,
    TimeGranularity,
)
from betfair_parser.spec.betting.type_definitions import (
    CompetitionResult,
    CountryCodeResult,
    EventResult,
    EventTypeResult,
    MarketBook,
    MarketCatalogue,
    MarketFilter,
    MarketProfitAndLoss,
    MarketTypeResult,
    PriceProjection,
    TimeRangeResult,
    VenueResult,
    betting_tag,
)
from betfair_parser.spec.common import (
    BetId,
    Date,
    EndpointType,
    Handicap,
    MarketId,
    Params,
    Request,
    Response,
    SelectionId,
    Set,
)


class _ListingParams(Params, frozen=True):
    filter: MarketFilter
    locale: str | None = None


class _ListingRequest(Request, kw_only=True, frozen=True, tag=betting_tag):
    endpoint_type = EndpointType.BETTING


class ListCompetitions(_ListingRequest, kw_only=True, frozen=True):
    """
    Returns a list of Competitions (i.e., World Cup 2013) associated with the markets selected by
    the MarketFilter. Currently only Football markets have an associated competition.
    """

    params: _ListingParams
    return_type = Response[list[CompetitionResult]]


class ListCountries(_ListingRequest, kw_only=True, frozen=True):
    """
    Returns a list of Countries associated with the markets selected by the MarketFilter.
    """

    params: _ListingParams
    return_type = Response[list[CountryCodeResult]]


class ListEvents(_ListingRequest, kw_only=True, frozen=True):
    """
    Returns a list of Events (i.e, Reading vs. Man United) associated with the markets selected
    by the MarketFilter.
    """

    params: _ListingParams
    return_type = Response[list[EventResult]]


class ListEventTypes(_ListingRequest, kw_only=True, frozen=True):
    """
    Returns a list of Event Types (i.e. Sports) associated with the markets selected by the
    MarketFilter.
    """

    params: _ListingParams
    return_type = Response[list[EventTypeResult]]


class ListMarketTypes(_ListingRequest, kw_only=True, frozen=True):
    """
    Returns a list of market types (i.e. MATCH_ODDS, NEXT_GOAL) associated with the markets selected
    by the MarketFilter. The market types are always the same, regardless of locale.
    """

    params: _ListingParams
    return_type = Response[list[MarketTypeResult]]


class ListVenues(_ListingRequest, kw_only=True, frozen=True):
    """
    Returns a list of Venues (i.e. Cheltenham, Ascot) associated with the markets selected by the
    MarketFilter. Currently, only Horse Racing markets are associated with a Venue.
    """

    params: _ListingParams
    return_type = Response[list[VenueResult]]


# More complex listings


class _ListMarketBookParams(Params, frozen=True):
    market_ids: Set[MarketId]  # NOTE: This is documented as list, but defined as set here for consistency
    price_projection: PriceProjection | None = None  # The desired projection of price data
    order_projection: OrderProjection | None = None  # The orders you want to receive in the response
    match_projection: MatchProjection | None = None  # If you ask for orders, specifies the representation of matches
    include_overall_position: bool | None = None  # If you ask for orders, returns matches for each selection
    partition_matched_by_strategy_ref: bool | None = None  # Breakdown of matches by strategy for each selection
    customer_strategy_refs: Set[str] | None = None
    currency_code: str | None = None  # A Betfair standard currency code
    locale: str | None = None  # The language used for the response
    matched_since: Date | None = None  # Only orders with at least one fragment matched since the specified date
    bet_ids: Set[BetId] | None = None  # Only orders with the specified bet IDs


class ListMarketBook(_ListingRequest, kw_only=True, frozen=True):
    """
    Returns a list of dynamic data about markets. Dynamic data includes prices, the status of the
    market, the status of selections, the traded volume, and the status of any orders you have
    placed in the market.

    Please note: Separate requests should be made for OPEN & CLOSED markets. Request that include
    both OPEN & CLOSED markets will only return those markets that are OPEN.

    Best Practice:
    Customers seeking to use listMarketBook to obtain price, volume, unmatched (EXECUTABLE) orders
    and matched position in a single operation should provide an OrderProjection of “EXECUTABLE” in
    their listMarketBook request and receive all unmatched (EXECUTABLE) orders and the aggregated
    matched volume from all orders irrespective of whether they are partially or fully matched.
    The level of matched volume aggregation (MatchProjection) requested should be
    ROLLED_UP_BY_AVG_PRICE or ROLLED_UP_BY_PRICE, the former being preferred. This provides a single
    call in which you can track prices, traded volume, unmatched orders and your evolving matched
    position with a reasonably fixed, minimally sized response.
    """

    params: _ListMarketBookParams
    return_type = Response[list[MarketBook]]


class _ListMarketCatalogueParams(Params, kw_only=True, frozen=True):
    filter: MarketFilter  # The filter to select desired markets
    market_projection: Set[MarketProjection] | None = None  # The type and amount of data returned about the market
    sort: MarketSort | None = None  # The order of the results, defaults to RANK
    max_results: int = 1000  # Limit on the total number of results returned
    locale: str | None = None  # The language used for the response


class ListMarketCatalogue(_ListingRequest, kw_only=True, frozen=True):
    """
    Returns a list of information about published (ACTIVE/SUSPENDED) markets that does not change
    (or changes very rarely). You use listMarketCatalogue to retrieve the name of the market, the
    names of selections and other information about markets. Market Data Request Limits apply to
    requests made to listMarketCatalogue.

    Please note: listMarketCatalogue does not return markets that are CLOSED.
    """

    params: _ListMarketCatalogueParams
    return_type = Response[list[MarketCatalogue]]


ListMarketCatalog = ListMarketCatalogue  # allow both spellings


class _ListMarketProfitAndLossParams(Params, frozen=True):
    market_ids: Set[MarketId]  # List of markets to calculate profit and loss
    include_settled_bets: bool | None = False  # Option to include settled bets (partially settled markets only)
    include_bsp_bets: bool | None = False  # Option to include BSP bets
    net_of_commission: bool | None = False  # Option to return profit and loss net of users current commission rate


class ListMarketProfitAndLoss(_ListingRequest, kw_only=True, frozen=True):
    """Returns a list of Countries associated with the markets selected by the MarketFilter."""

    params: _ListMarketProfitAndLossParams
    return_type = Response[list[MarketProfitAndLoss]]


class _ListRunnerBookParams(Params, frozen=True):
    market_id: MarketId  # The unique id for the market
    selection_id: SelectionId  # The unique id for the selection in the market
    handicap: Handicap | None = None  # The handicap associated with the runner in case of Asian handicap market
    price_projection: PriceProjection | None = None
    order_projection: OrderProjection | None = None
    match_projection: MatchProjection | None = None  # If you ask for orders, specifies the representation of matches
    include_overall_position: bool | None = None  # If you ask for orders, returns matches for each selection
    partition_matched_by_strategy_ref: bool | None = None  # Return a breakdown of matches by strategy
    customer_strategy_refs: Set[str] | None = None
    currency_code: str | None = None  # A Betfair standard currency code
    locale: str | None = None  # The language used for the response
    matched_since: Date | None = None  # Restricts to orders with at least one fragment matched since specified date
    bet_ids: Set[BetId] | None = None  # Restricts to orders with the specified bet IDs


class ListRunnerBook(_ListingRequest, kw_only=True, frozen=True):
    """
    Returns a list of dynamic data about a market and a specified runner. Dynamic data includes
    prices, the status of the market, the status of selections, the traded volume, and the status
    of any orders you have placed in the market.
    """

    params: _ListRunnerBookParams
    return_type = Response[list[MarketBook]]


class _ListTimeRangesParams(Params, frozen=True):
    # The filter to select desired markets. All markets that match the criteria in the filter are selected.
    filter: MarketFilter
    # The granularity of time periods that correspond to markets selected by the market filter.
    granularity: TimeGranularity


class ListTimeRanges(_ListingRequest, kw_only=True, frozen=True):
    """
    Returns a list of time ranges in the granularity specified in the request (i.e. 3PM to 4PM,
    Aug 14th to Aug 15th) associated with the markets selected by the MarketFilter.
    """

    params: _ListTimeRangesParams
    return_type = Response[list[TimeRangeResult]]



"""
Minimalistic client example

`session` should be some requests session or httpx client.

This library aims for compatibility with any kind of http client. Thus, the objects within
the specification provide a straight forward way to quickly construct a http request out of
a prepared header and body.
"""

from betfair_parser.endpoints import ENDPOINTS
from betfair_parser.spec.common import Request
from betfair_parser.spec.identity import CertLogin, KeepAlive, Login, Logout


def request(session, req: Request, endpoints=ENDPOINTS):
    """Minimalistic client example."""

    url = endpoints.url_for_request(req)
    raw_resp = session.post(url, headers=req.headers(), data=req.body())
    raw_resp.raise_for_status()
    return req.parse_response(raw_resp.content, raise_errors=True)


def login(session, username, password, app_key, two_factor_code="", endpoints=ENDPOINTS):
    """Authenticate a session to betfair."""

    session.headers.update({"X-Application": app_key})
    resp = request(
        session,
        Login.with_params(username=username, password=password + two_factor_code),
        endpoints=endpoints,
    )
    session.headers.update(
        {
            "X-Authentication": resp.token,
            "X-Application": resp.product,
        }
    )


def keep_alive(session, endpoints=ENDPOINTS):
    """Renew authentication."""

    resp = request(session, KeepAlive(), endpoints=endpoints)
    session.headers.update({"X-Authentication": resp.token})


def cert_login(session, username, password, app_key, endpoints=ENDPOINTS):
    """Bot authentication. Session certificates need to be properly configured already."""

    session.headers.update({"X-Application": app_key})
    resp = request(
        session,
        CertLogin.with_params(username=username, password=password),
        endpoints=endpoints,
    )
    session.headers.update({"X-Authentication": resp.token})


def logout(session, endpoints=ENDPOINTS):
    try:
        request(session, Logout(), endpoints=endpoints)
    finally:
        session.headers.pop("X-Authentication", None)



"""
Definition of the betfair streaming API messages as defined in:
- https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Exchange+Stream+API
- https://github.com/betfair/stream-api-sample-code/blob/master/ESASwaggerSchema.json
"""

from typing import Literal

from betfair_parser.spec.common import BaseMessage, first_lower
from betfair_parser.spec.streaming.enums import ChangeType, SegmentType, StatusErrorCode
from betfair_parser.spec.streaming.type_definitions import (
    MarketChange,
    MarketDataFilter,
    MarketFilter,
    OrderFilter,
    OrderMarketChange,
    StreamRef,
)


class _StreamRequest(BaseMessage, tag_field="op", tag=first_lower, frozen=True):
    """Common parent class for any stream request."""

    id: StreamRef | None = None  # Client generated unique id to link request with response (like json rpc)


class _StreamResponse(BaseMessage, tag_field="op", tag=str.lower, frozen=True):
    """Common parent class for any stream response."""

    id: StreamRef | None = None  # Client generated unique id to link request with response (like json rpc)


class Authentication(_StreamRequest, kw_only=True, frozen=True):
    """
    This message is the first message that the client must send on connecting to the server. You must
    be authenticated before any other request is processed.
    """

    app_key: str  # Application Key
    session: str  # Session Token generated by the API login


class _Subscription(_StreamRequest, kw_only=True, frozen=True):
    """Common parent class for any Subscription request."""

    clk: str | None = None  # Token value delta (received in MarketChangeMessage) for resuming a subscription
    conflate_ms: int | None = None  # The conflation rate (looped back on initial image: bounds are 0 to 120000)
    heartbeat_ms: int | None = None  # The heartbeat rate (looped back on initial image: bounds are 500 to 5000)
    initial_clk: str | None = None  # Token value that should be passed to resume a subscription
    segmentation_enabled: bool = True  # allow server to send large sets of data in segments, instead of a single block


class MarketSubscription(_Subscription, kw_only=True, frozen=True):
    """
    This subscription type is used to receive price changes for one or more markets; your subscription
    criteria determine what you see. Limiting the amount of data that you consume will make your initial
    image much smaller (and faster) & suppress changes that are uninteresting to you.
    """

    market_filter: MarketFilter
    market_data_filter: MarketDataFilter


class OrderSubscription(_Subscription, kw_only=True, frozen=True):
    """This subscription type is used to receive order changes."""

    order_filter: OrderFilter | None = None  # Optional filter applied to order subscription


class Heartbeat(_StreamRequest, frozen=True):
    """
    This is an explicit heartbeat request (in addition to the server heartbeat interval which is automatic).
    This functionality should not normally be necessary unless you need to keep a firewall open.
    """


class Connection(_StreamResponse, kw_only=True, frozen=True):
    """This is received by the client when it successfully opens a connection to the server."""

    connection_id: str  # Unique identifier for support queries


class Status(_StreamResponse, kw_only=True, frozen=True):
    """Every request receives a status response with a matching id."""

    connection_closed: bool
    connection_id: str | None = None
    connections_available: int | None = None  # The number of connections available for this account at this moment
    error_code: StatusErrorCode | None = None
    error_message: str | None = None  # Additional message in case of a failure
    status_code: Literal["SUCCESS", "FAILURE"]  # The status of the last request

    @property
    def is_error(self):
        return self.status_code != "SUCCESS"


class _ChangeMessage(_StreamResponse, kw_only=True, frozen=True):
    """Common parent class for any ChangeMessage."""

    clk: str | None = None  # Token value (non-None) should be stored for resuming in case of a disconnect
    conflate_ms: int | None = None  # The conflation rate (may differ from that requested if subscription is delayed)
    ct: ChangeType | None = None
    heartbeat_ms: int | None = None  # Heartbeat rate (may differ from requested: bounds are 500 to 30000)
    initial_clk: str | None = None  # Token value (non-None) should be stored for resuming in case of a disconnect
    pt: int  # Publish Time
    segment_type: SegmentType | None = None  # denotes the beginning and end of a segmentation
    status: int | None = None  # None if the stream is up-to-date and 503 if the services are experiencing latencies

    @property
    def is_heartbeat(self):
        return self.ct == ChangeType.HEARTBEAT

    @property
    def stream_unreliable(self):
        return self.status == 503

    @property
    def change_type(self):
        """Set to indicate the type of change (if None this is a delta)"""
        return self.ct

    @property
    def publish_time(self):
        """Time (in milliseconds since epoch) that the changes were generated"""
        return self.pt


class MCM(_ChangeMessage, kw_only=True, frozen=True):
    """
    This is the ChangeMessage stream of data Betfair sends back after subscribing to the market stream.
    Market subscriptions are always in the underlying exchange currency - GBP.
    """

    mc: list[MarketChange] | None = None

    @property
    def market_changes(self):
        """The modifications to markets (will be None on a heartbeat)"""
        return self.mc


class OCM(_ChangeMessage, kw_only=True, frozen=True):
    """
    This is the ChangeMessage stream of data Betfair sends back after subscribing to the order stream.
    Order subscriptions are provided in the currency of the account that the orders are placed in.
    """

    oc: list[OrderMarketChange] | None = None

    @property
    def order_market_changes(self):
        """The modifications to account's orders (will be None on a heartbeat)"""
        return self.oc



import datetime
from typing import Annotated, TypeAlias

import msgspec

from betfair_parser.spec.common.messages import BaseMessage


Date = Annotated[datetime.datetime, msgspec.Meta(title="Date", tz=True)]
IDType = Annotated[
    int, msgspec.Meta(title="IDType", description="integer data, but defined and encoded as string", ge=0)
]

# Betfair defines several input fields as set. Nevertheless, the JSON data is encoded as list. For convenience,
# these data fields should accept a set or a list as input. But currently, there seems to be no good
# solution, to satisfy as well the type checkers as the restrictions of msgspec. Sticking with set
# seems to be the most correct option for now.

Set: TypeAlias = set

# Type aliases as defined within the XML specification with minimalistic validation added.

MarketType = Annotated[str, msgspec.Meta(title="MarketType")]
Venue = Annotated[str, msgspec.Meta(title="Venue")]
MarketId = Annotated[str, msgspec.Meta(title="MarketId")]  # The only ID, that actually needs to be a string
SelectionId = Annotated[int, msgspec.Meta(title="SelectionId", ge=0)]  # The only ID, that is actually defined as int
Handicap = Annotated[float, msgspec.Meta(title="Handicap", gt=-1000, lt=1000)]
EventId = Annotated[IDType, msgspec.Meta(title="EventId")]
EventTypeId = Annotated[IDType, msgspec.Meta(title="EventTypeId")]
CountryCode = Annotated[str, msgspec.Meta(title="CountryCode", min_length=2, max_length=3)]
ExchangeId = Annotated[IDType, msgspec.Meta(title="ExchangeId")]
CompetitionId = Annotated[IDType, msgspec.Meta(title="CompetitionId")]
Price = Annotated[float, msgspec.Meta(title="Price", ge=0, le=1001)]
Size = Annotated[float, msgspec.Meta(title="Size", ge=0)]
BetId = Annotated[IDType, msgspec.Meta(title="BetId")]
MatchId = Annotated[IDType, msgspec.Meta(title="MatchId")]
CustomerRef = Annotated[str, msgspec.Meta(title="CustomerRef")]
CustomerOrderRef = Annotated[str, msgspec.Meta(title="CustomerOrderRef")]
CustomerStrategyRef = Annotated[str, msgspec.Meta(title="CustomerStrategyRef")]


class TimeRange(BaseMessage, frozen=True):
    from_: Date | None = msgspec.field(name="from", default=None)
    to: Date | None = None
